import os
import logging
from collections import defaultdict
from time import time
import torch
import mne
import numpy as np
import pandas as pd
from mne.time_frequency.multitaper import psd_array_multitaper
from scipy.io import loadmat
from scipy.signal import welch

LOG = logging.getLogger("meegnet")


def extract_bands(data: np.array, f: list = None) -> np.array:
    """extract_bands.

    Parameters
    ----------
    data : np.array
        the data after it has been transformed to frequency space. Of shape n_samples x n_channels x n_bins
        or n_channels x n_bins
    f : list
        the list of bins. id set to None, a list of all bins every .5 from 0 to n_bins will be generated.

    Returns
    -------
    data : np.array
        the data after averaging for frequency bands of shape n_samples x n_channels x 7 or n_channels x 7 depending
        on input shape.
        bands are defined as follow:
            delta: .5 to 4 Hz
            theta: 4 to 8 Hz
            alpha: 8 to 12 Hz
            beta1: 12 to 30 Hz
            beta2: 30 to 60 Hz
            gamma1: 60 to 90 Hz
            gamma2: 90 to 120 Hz

    """
    if f is None:
        f = np.asarray([float(i / 2) for i in range(data.shape[-1])])
    data = [
        data[..., (f >= 0.5) * (f <= 4)].mean(axis=-1)[..., None],
        data[..., (f >= 4) * (f <= 8)].mean(axis=-1)[..., None],
        data[..., (f >= 8) * (f <= 12)].mean(axis=-1)[..., None],
        data[..., (f >= 12) * (f <= 30)].mean(axis=-1)[..., None],
        data[..., (f >= 30) * (f <= 60)].mean(axis=-1)[..., None],
        data[..., (f >= 60) * (f <= 90)].mean(axis=-1)[..., None],
        data[..., (f >= 90) * (f <= 120)].mean(axis=-1)[..., None],
    ]
    data = np.concatenate(data, axis=-1)
    return data


def compute_psd(data: np.array, fs: int, option: str = "multitaper"):
    """compute_psd.

    Parameters
    ----------
    data : np.array
        The data to compute psd on. Must be of shape n_channels x n_samples.
    fs : int
        The sampling frequency.
    option : str
        method option for the psd computation. Can be welch or multitaper.
    """
    """"""
    mne.set_log_level(verbose=False)
    if option == "multitaper":
        psd, f = psd_array_multitaper(data, sfreq=fs, fmax=150)
    elif option == "welch":
        f, psd = welch(data, fs=fs)
    else:
        raise "Error: invalid option for psd computation."
    return extract_bands(psd, f)


def cuda_check():
    """
    Checks if a CUDA device is available and returns it.

    Returns
    -------
    torch.device
        A CUDA device if one is available, otherwise a CPU device.
    """
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if device.type == "cuda":
        LOG.info("CUDA is available! Using GPU.")
    else:
        LOG.info("CUDA is not available. Using CPU.")
    return device


def check_PD(mat):
    if len(mat.shape) > 2:
        out = []
        for submat in mat:
            out.append(check_PD(submat))
        return np.array(out)
    else:
        return np.all(np.linalg.eigvals(mat) > 0)


def strip_string(string: str) -> str:
    return string.translate({ord(i): None for i in "!@#$[]'"})


def load_psd_cc_subjects(PSD_PATH, sub_info_path, window, overlap):
    df = pd.read_csv(sub_info_path)
    sub_list = df["Observations"].tolist()
    labels = list(df["gender_code"] - 1)
    psd = []
    for sub in sub_list:
        file_path = os.path.join(PSD_PATH, "PSD_{}_{}_{}".format(sub, window, overlap))
        try:
            psd.append(loadmat(file_path)["data"].ravel())
        except IOError:
            LOG.info(sub, "Not Found")
    return np.array(psd), np.array(labels)


def nice_time(time):
    """Returns time in a humanly readable format."""
    m, h, j = 60, 3600, 24 * 3600
    nbj = time // j
    nbh = (time - j * nbj) // h
    nbm = (time - j * nbj - h * nbh) // m
    nbs = time - j * nbj - h * nbh - m * nbm
    if time > m:
        if time > h:
            if time > j:
                nt = "%ij, %ih:%im:%is" % (nbj, nbh, nbm, nbs)
            else:
                nt = "%ih:%im:%is" % (nbh, nbm, nbs)
        else:
            nt = "%im:%is" % (nbm, nbs)
    elif time < 1:
        nt = "<1s"
    else:
        nt = "%is" % nbs
    return nt


def elapsed_time(t0, t1):
    """Time lapsed between t0 and t1.

    Returns the time (from time.time()) between t0 and t1 in a
    more readable fashion.

    Parameters
    ----------
    t0: float,
        time.time() initial measure of time
        (eg. at the begining of the script)
    t1: float,
        time.time() time at the end of the script
        or the execution of a function.

    """
    lapsed = abs(t1 - t0)
    return nice_time(lapsed)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Things that don't belong anywhere else
"""

import hashlib
import json
import os
import sys
from shutil import copyfile
from collections import OrderedDict
from numbers import Number
import operator

import numpy as np
import torch
import tqdm
from collections import Counter

def make_weights_for_balanced_classes(dataset):
    counts = Counter()
    classes = []
    for _, y in dataset:
        y = int(y)
        counts[y] += 1
        classes.append(y)

    n_classes = len(counts)

    weight_per_class = {}
    for y in counts:
        weight_per_class[y] = 1 / (counts[y] * n_classes)

    weights = torch.zeros(len(dataset))
    for i, y in enumerate(classes):
        weights[i] = weight_per_class[int(y)]

    return weights

def pdb():
    sys.stdout = sys.__stdout__
    import pdb
    print("Launching PDB, enter 'n' to step to parent function.")
    pdb.set_trace()

def seed_hash(*args):
    """
    Derive an integer hash from all args, for use as a random seed.
    """
    args_str = str(args)
    return int(hashlib.md5(args_str.encode("utf-8")).hexdigest(), 16) % (2**31)

def print_separator():
    print("="*80)

def print_row(row, colwidth=10, latex=False):
    if latex:
        sep = " & "
        end_ = "\\\\"
    else:
        sep = "  "
        end_ = ""

    def format_val(x):
        if np.issubdtype(type(x), np.floating):
            x = "{:.10f}".format(x)
        return str(x).ljust(colwidth)[:colwidth]
    print(sep.join([format_val(x) for x in row]), end_)

class _SplitDataset(torch.utils.data.Dataset):
    """Used by split_dataset"""
    def __init__(self, underlying_dataset, keys):
        super(_SplitDataset, self).__init__()
        self.underlying_dataset = underlying_dataset
        self.keys = keys
    def __getitem__(self, key):
        return self.underlying_dataset[self.keys[key]]
    def __len__(self):
        return len(self.keys)

def split_dataset(dataset, n, seed=0):
    """
    Return a pair of datasets corresponding to a random split of the given
    dataset, with n datapoints in the first dataset and the rest in the last,
    using the given random seed
    """
    assert(n <= len(dataset))
    keys = list(range(len(dataset)))
    np.random.RandomState(seed).shuffle(keys)
    keys_1 = keys[:n]
    keys_2 = keys[n:]
    return _SplitDataset(dataset, keys_1), _SplitDataset(dataset, keys_2)

def random_pairs_of_minibatches(minibatches):
    perm = torch.randperm(len(minibatches)).tolist()
    pairs = []

    for i in range(len(minibatches)):
        j = i + 1 if i < (len(minibatches) - 1) else 0

        xi, yi = minibatches[perm[i]][0], minibatches[perm[i]][1]
        xj, yj = minibatches[perm[j]][0], minibatches[perm[j]][1]

        min_n = min(len(xi), len(xj))

        pairs.append(((xi[:min_n], yi[:min_n]), (xj[:min_n], yj[:min_n])))

    return pairs

def accuracy(network, loader, weights, device):
    correct = 0
    total = 0
    weights_offset = 0

    network.eval()
    with torch.no_grad():
        for x, y in loader:
            x = x.to(device)
            y = y.to(device)
            p = network.predict(x)
            if weights is None:
                batch_weights = torch.ones(len(x))
            else:
                batch_weights = weights[weights_offset : weights_offset + len(x)]
                weights_offset += len(x)
            batch_weights = batch_weights.to(device)
            if p.size(1) == 1:
                correct += (p.gt(0).eq(y).float() * batch_weights.view(-1, 1)).sum().item()
            else:
                correct += (p.argmax(1).eq(y).float() * batch_weights).sum().item()
            total += batch_weights.sum().item()
    network.train()

    return correct / total

class Tee:
    def __init__(self, fname, mode="a"):
        self.stdout = sys.stdout
        self.file = open(fname, mode)

    def write(self, message):
        self.stdout.write(message)
        self.file.write(message)
        self.flush()

    def flush(self):
        self.stdout.flush()
        self.file.flush()

class ParamDict(OrderedDict):
    """Code adapted from https://github.com/Alok/rl_implementations/tree/master/reptile.
    A dictionary where the values are Tensors, meant to represent weights of
    a model. This subclass lets you perform arithmetic on weights directly."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, *kwargs)

    def _prototype(self, other, op):
        if isinstance(other, Number):
            return ParamDict({k: op(v, other) for k, v in self.items()})
        elif isinstance(other, dict):
            return ParamDict({k: op(self[k], other[k]) for k in self})
        else:
            raise NotImplementedError

    def __add__(self, other):
        return self._prototype(other, operator.add)

    def __rmul__(self, other):
        return self._prototype(other, operator.mul)

    __mul__ = __rmul__

    def __neg__(self):
        return ParamDict({k: -v for k, v in self.items()})

    def __rsub__(self, other):
        # a- b := a + (-b)
        return self.__add__(other.__neg__())

    __sub__ = __rsub__

    def __truediv__(self, other):
        return self._prototype(other, operator.truediv)



import os
from collections import defaultdict, OrderedDict
import logging
import torch
from torch import nn, optim
from torch.autograd import Variable
import torch.nn.functional as F
from torch.utils.data import DataLoader
from scipy.io import loadmat, savemat
import numpy as np
from huggingface_hub import hf_hub_download

LOG = logging.getLogger("meegnet")


def create_net(net_option, input_size, n_outputs, net_params=None):
    if net_option == "MLP":
        return MLP(
            input_size=input_size,
            n_outputs=n_outputs,
            hparams={
                "mlp_width": net_params["linear"],
                "mlp_depth": net_params["hlayers"],
                "mlp_dropout": net_params["dropout"],
            },
        )
    elif net_option == "meegnet":
        return my_net(input_size, n_outputs)
    elif net_option == "custom_net":
        return FullNet(
            input_size,
            n_outputs,
            net_params["hlayers"],
            net_params["filters"],
            net_params["nchan"],
            net_params["linear"],
            net_params["dropout"],
            net_params["batchnorm"],
            net_params["maxpool"],
        )
    elif net_option == "VGG":
        return VGG16_NET(input_size, n_outputs)
    elif net_option == "EEGNet":
        return EEGNet(input_size, n_outputs)
    elif net_option == "vanPutNet":
        return vanPutNet(input_size, n_outputs)


class Flatten(nn.Module):
    # Flatten layer used to connect between feature extraction and classif parts of a net.
    def forward(self, x):
        x = x.view(x.size(0), -1)
        return x


class DepthwiseConv2d(nn.Module):
    def __init__(self, in_channels, kernel_size, depthwise_multiplier=1, **kwargs):
        super(DepthwiseConv2d, self).__init__()
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels * depthwise_multiplier,
            kernel_size,
            groups=in_channels,
            **kwargs,
        )

    def forward(self, x):
        return self.depthwise(x)


class SeparableConv2d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, **kwargs):
        super(SeparableConv2d, self).__init__()
        self.depthwise = DepthwiseConv2d(in_channels, kernel_size, **kwargs)
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1, **kwargs)

    def forward(self, x):
        return self.pointwise(self.depthwise(x))


class customNet(nn.Module):
    def __init__(self, input_size, n_outputs):
        super(customNet, self).__init__()
        self.input_size = input_size
        self.n_outputs = n_outputs

    def forward(self, x):
        feats = self.feature_extraction(x)
        outs = self.classif(feats)
        return outs

    def _get_lin_size(self, layers):
        return nn.Sequential(*layers)(torch.zeros((1, *self.input_size))).shape[-1]


class EEGNet(customNet):
    def __init__(
        self,
        input_size,
        n_outputs,
        filter_size=64,
        n_filters=16,
        dropout=0.5,
        dropout_option="Dropout",
        depthwise_multiplier=2,
    ):
        customNet.__init__(self, input_size, n_outputs)
        if dropout_option == "SpatialDropout2D":
            dropoutType = nn.Dropout2d
        elif dropout_option == "Dropout":
            dropoutType = nn.Dropout
        else:
            raise ValueError(
                "dropoutType must be one of SpatialDropout2D "
                "or Dropout, passed as a string."
            )

        n_channels = input_size[1]
        layer_list = [
            nn.Conv2d(
                input_size[0],
                n_filters,
                (1, filter_size),
                padding=(1, (filter_size) // 2),
                # padding="same",
                bias=False,
            ),
            nn.BatchNorm2d(n_filters),
            # depthwise_constraind=maxnorm(1.) not used
            DepthwiseConv2d(
                n_filters,
                (n_channels, 1),
                depthwise_multiplier=depthwise_multiplier,
                padding=0,
                bias=False,
            ),
            # Changed n_filters to *2 becaus of dimension error,
            # TODO check if it was originally a typo in our code
            nn.BatchNorm2d(n_filters * 2),
            nn.ELU(),
            nn.AvgPool2d((1, 4)),
            dropoutType(dropout),
            SeparableConv2d(n_filters * 2, n_filters * 2, (1, 16), padding=(1, 8), bias=False),
            nn.BatchNorm2d(n_filters * 2),
            nn.ELU(),
            nn.AvgPool2d((1, 8)),
            dropoutType(dropout),
            Flatten(),
        ]

        layers = nn.ModuleList(layer_list)
        lin_size = self._get_lin_size(layers)

        self.feature_extraction = nn.Sequential(*layers)

        self.classif = nn.Sequential(
            *nn.ModuleList(
                [
                    # not using the kernel_constraint=max_norm(norm_rate) parameter
                    nn.Linear(lin_size, n_outputs),
                ]
            )
        )

    def forward(self, x):
        feats = self.feature_extraction(x)
        outs = self.classif(feats)
        return outs


# This implementation is rather common and found on various blogs/github repos
class VGG16_NET(customNet):
    def __init__(self, input_size, n_outputs):
        super(VGG16_NET, self).__init__(input_size, n_outputs)

        self.maxpool = nn.MaxPool2d(kernel_size=2, stride=2)
        layer_list = [
            nn.Conv2d(in_channels=3, out_channels=64, kernel_size=3, padding=1),
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, padding=1),
            self.maxpool,
            nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, padding=1),
            nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, padding=1),
            self.maxpool,
            nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, padding=1),
            nn.Conv2d(in_channels=256, out_channels=256, kernel_size=3, padding=1),
            nn.Conv2d(in_channels=256, out_channels=256, kernel_size=3, padding=1),
            self.maxpool,
            nn.Conv2d(in_channels=256, out_channels=512, kernel_size=3, padding=1),
            nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, padding=1),
            nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, padding=1),
            self.maxpool,
            nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, padding=1),
            nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, padding=1),
            nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, padding=1),
            self.maxpool,
            Flatten(),
        ]
        layers = nn.ModuleList(layer_list)
        self.feature_extraction = nn.Sequential(*layers)

        lin_size = self._get_lin_size(layers)
        self.classif = nn.Sequential(
            *nn.ModuleList(
                [
                    nn.Linear(lin_size, 4096),
                    nn.ReLU(),
                    nn.Dropout(),
                    nn.Linear(4096, 4096),
                    nn.ReLU(),
                    nn.Dropout(),
                    nn.Linear(4096, self.n_outputs),
                ]
            )
        )


class MLP(customNet):
    """Just an MLP"""

    def __init__(self, input_size, n_outputs, hparams):
        customNet.__init__(self, input_size, n_outputs)
        n_inputs = np.prod(input_size)
        self.flatten = Flatten()
        self.input = nn.Linear(n_inputs, hparams["mlp_width"])
        self.dropout = nn.Dropout(hparams["mlp_dropout"])
        self.hiddens = nn.ModuleList(
            [
                nn.Linear(hparams["mlp_width"], hparams["mlp_width"])
                for _ in range(hparams["mlp_depth"] - 2)
            ]
        )
        self.output = nn.Linear(hparams["mlp_width"], n_outputs)
        self.n_outputs = n_outputs

    def forward(self, x):
        x = self.flatten(x)
        x = self.input(x)
        x = self.dropout(x)
        x = F.relu(x)
        for hidden in self.hiddens:
            x = hidden(x)
            x = self.dropout(x)
            x = F.relu(x)
        x = self.output(x)
        return x


class my_net(customNet):
    def __init__(
        self,
        input_size,
        n_outputs,
        n_linear=2000,
        dropout=0.5,
    ):
        super(my_net, self).__init__(input_size, n_outputs)
        self.maxpool = nn.MaxPool2d(kernel_size=(1, 20), stride=1)
        layer_list = [
            nn.Conv2d(input_size[0], 100, (input_size[1], 1)),
            nn.ReLU(),
            nn.Conv2d(100, 200, (1, 9)),
            self.maxpool,
            nn.ReLU(),
            nn.Conv2d(200, 200, (1, 9)),
            self.maxpool,
            nn.ReLU(),
            nn.Conv2d(200, 100, (1, 9)),
            self.maxpool,
            nn.ReLU(),
            Flatten(),
            nn.Dropout(dropout),
        ]

        layers = nn.ModuleList(layer_list)
        lin_size = self._get_lin_size(layers)

        self.feature_extraction = nn.Sequential(*layers)
        self.classif = nn.Sequential(
            *nn.ModuleList(
                [
                    nn.Linear(lin_size, int(n_linear / 2)),
                    nn.Linear(int(n_linear / 2), n_outputs),
                ]
            )
        )


class FullNet(nn.Module):
    def __init__(
        self,
        input_size,
        n_outputs,
        hlayers=2,
        filter_size=7,
        nchan=5,
        n_linear=150,
        dropout=0.25,
        batchnorm=False,
        maxpool=0,
    ):
        super(FullNet, self).__init__()
        self.input_size = input_size

        layer_list = [
            nn.Conv2d(input_size[0], nchan, (input_size[1], 1)),
            nn.ReLU(),
        ]
        prev = nchan
        for i in range(0, int(hlayers / 2)):
            nex = prev * 2
            layer_list += [nn.Conv2d(prev, nex, (1, filter_size))]
            if batchnorm:
                layer_list += [nn.BatchNorm2d(nex)]
            if maxpool != 0:
                layer_list += [nn.MaxPool2d((1, maxpool), 1)]
            layer_list += [nn.ReLU()]
            prev = nex

        if hlayers % 2 != 0:
            layer_list += [nn.Conv2d(prev, prev, (1, filter_size))]
            if batchnorm:
                layer_list += [nn.BatchNorm2d(prev)]
            if maxpool != 0:
                layer_list += [nn.MaxPool2d((1, maxpool), 1)]
            layer_list += [nn.ReLU()]

        for i in range(0, int(hlayers / 2)):
            nex = int(prev / 2)
            layer_list += [nn.Conv2d(prev, nex, (1, filter_size))]
            if batchnorm:
                layer_list += [nn.BatchNorm2d(nex)]
            if maxpool != 0:
                layer_list += [nn.MaxPool2d((1, maxpool), 1)]
            layer_list += [nn.ReLU()]
            prev = nex

        layer_list += [
            Flatten(),
            nn.Dropout(dropout),
        ]

        layers = nn.ModuleList(layer_list)
        lin_size = self._get_lin_size(layers)

        self.feature_extraction = nn.Sequential(*layers)
        self.classif = nn.Sequential(
            *nn.ModuleList(
                [
                    nn.Linear(lin_size, int(n_linear / 2)),
                    nn.Linear(int(n_linear / 2), n_outputs),
                ]
            )
        )

    def forward(self, x):
        feats = self.feature_extraction(x)
        outs = self.classif(feats)
        return outs

    def _get_lin_size(self, layers):
        return nn.Sequential(*layers)(torch.zeros((1, *self.input_size))).shape[-1]


class vanPutNet(customNet):
    def __init__(self, input_size, n_output, dropout=0.25):
        customNet.__init__(self, input_size, n_output)
        layers = nn.ModuleList(
            [
                nn.Conv2d(1, 100, 3),
                nn.ReLU(),
                nn.MaxPool2d((2, 2)),
                nn.Dropout(dropout),
                nn.Conv2d(100, 100, 3),
                nn.MaxPool2d((2, 2)),
                nn.Dropout(dropout),
                nn.Conv2d(100, 300, (2, 3)),
                nn.MaxPool2d((2, 2)),
                nn.Dropout(dropout),
                nn.Conv2d(300, 300, (1, 7)),
                nn.MaxPool2d((2, 2)),
                nn.Dropout(dropout),
                nn.Conv2d(300, 100, (1, 3)),
                nn.Conv2d(100, 100, (1, 3)),
                Flatten(),
            ]
        )

        lin_size = self._get_lin_size(layers)
        layers.append(nn.Linear(lin_size, 2))
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class AutoEncoder(customNet):
    def __init__(
        self,
        input_size,
    ):
        customNet.__init__(self, input_size)

        lin_size = input_size[0] * input_size[1] * input_size[2]

        self.encoder = nn.Sequential(
            nn.Linear(lin_size, 4096),
            nn.ReLU(True),
            nn.Linear(4096, 2048),
        )

        self.decoder = nn.Sequential(
            nn.Linear(2048, 4096),
            nn.ReLU(True),
            nn.Linear(4096, lin_size),
            nn.Tanh(),
        )

    def forward(self, x):
        x = self.encoder(x)
        return self.decoder(x)


class Model:
    def __init__(
        self,
        name,
        net_option,
        input_size,
        n_outputs,
        save_path=None,
        learning_rate=0.00001,
        optimizer=optim.Adam,
        criterion=nn.CrossEntropyLoss(),
        n_folds=5,
        device="cuda",
    ):
        self.name = name
        self.input_size = input_size  # TODO here put assertions on the shape
        self.net = create_net(net_option, input_size, n_outputs)
        self.n_outputs = n_outputs
        self.n_folds = n_folds
        self.criterion = criterion
        self.save_path = save_path
        self.lr = learning_rate
        self.optimizer = optimizer(self.net.parameters(), lr=learning_rate)
        self.results = defaultdict(lambda: 0)
        self.checkpoint = defaultdict(lambda: 0)

        if torch.cuda.is_available():
            self.device = "cuda"
        elif device == "cuda":
            LOG.warning("Warning: gpu device requested but unavailable. Setting device to CPU")
            self.device = "cpu"
        self.net.to(self.device)

    def train(
        self,
        dataset,
        batch_size=128,
        patience=20,
        max_epoch=None,
        model_path=None,
        num_workers=4,
        load=False,
    ):
        train_accs = []
        valid_accs = []
        train_losses = []
        valid_losses = []
        best_vloss = float("inf")
        best_vacc = 0.5
        best_epoch = 0
        patience_state = self.results["current_patience"]
        epoch = self.checkpoint["epoch"]
        self.net.train()
        self.batch_size = batch_size
        self.num_workers = num_workers

        LOG.info("Creating DataLoaders...")
        train_index, valid_index, test_index = dataset.data_split(0.8, 0.1, 0.1)
        trainloader = DataLoader(
            dataset.torchDataset(train_index),
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
            pin_memory=True,
        )
        validloader = DataLoader(
            dataset.torchDataset(valid_index),
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
            pin_memory=True,
        )

        if load:
            LOG.info("Loading previous network state...")
            self.load(model_path)

        LOG.info("Starting Training with:")
        LOG.info(f"batch size: {batch_size}")
        LOG.info(f"learning rate: {self.lr}")
        LOG.info(f"patience: {patience}")
        while patience_state < patience:
            epoch += 1
            n_batches = len(trainloader)
            for i, batch in enumerate(trainloader):
                self.optimizer.zero_grad()
                if len(batch) > 2:
                    X, y, groups = batch
                else:
                    X, y = batch
                y = y.view(-1).to(self.device)
                X = X.view(-1, *self.input_size).to(self.device)
                out = self.net.forward(X.float())
                loss = self.criterion(out, Variable(y.long()))
                loss.backward()
                self.optimizer.step()
                progress = f"Epoch: {epoch} // Batch {i+1}/{n_batches} // loss = {loss:.5f}"
                if n_batches > 10:
                    if i % (n_batches // 10) == 0:
                        LOG.info(progress)
                else:
                    LOG.info(progress)
            train_loss, train_acc = self.evaluate(trainloader)
            valid_loss, valid_acc = self.evaluate(validloader)
            train_accs.append(train_acc)
            valid_accs.append(valid_acc)
            train_losses.append(train_loss)
            valid_losses.append(valid_loss)

            if valid_loss < best_vloss:
                best_vacc = valid_acc
                best_vloss = valid_loss
                best_epoch = epoch
                patience_state = 0
                self.results = {
                    "acc_score": [best_vacc],
                    "loss_score": [best_vloss],
                    "acc": valid_accs,
                    "train_acc": train_accs,
                    "valid_loss": valid_losses,
                    "train_loss": train_losses,
                    "best_epoch": best_epoch,
                    "patience": patience,
                    "current_patience": patience_state,
                }
                self.checkpoint = {
                    "epoch": epoch + 1,
                    "state_dict": self.net.state_dict(),
                    "optimizer": self.optimizer.state_dict(),
                }
                if self.save_path is not None:
                    self.save(model_path)
            else:
                patience_state += 1
            LOG.info("Epoch: {}".format(epoch))
            LOG.info(" [LOSS] TRAIN {} / VALID {}".format(train_loss, valid_loss))
            LOG.info(" [ACC] TRAIN {} / VALID {}".format(train_acc, valid_acc))
            if max_epoch is not None:
                if epoch == max_epoch:
                    LOG.info("Max number of epoch reached. Stopping training.")
                    break

    def fit(self, *args, **kwargs):
        return self.train(args, *kwargs)

    def evaluate(self, dataloader):
        with torch.no_grad():
            losses = 0
            accuracy = 0
            counter = 0
            for batch in dataloader:
                if len(batch) > 2:
                    X, y, _ = batch
                else:
                    X, y = batch
                y = y.view(-1).to(self.device)
                X = X.view(-1, *self.input_size).to(self.device)
                out = self.net.forward(X.float())
                loss = self.criterion(out, Variable(y.long()))
                acc = self.compute_accuracy(out, y)
                n = y.size(0)
                losses += loss.detach().sum().data.cpu().numpy() * n
                accuracy += acc.sum().data.cpu().numpy() * n
                counter += n
            floss = losses / float(counter)
            faccuracy = accuracy / float(counter)
        return floss, faccuracy

    def test(self, dataset):
        _, _, test_index = dataset.data_split(0.8, 0.1, 0.1)
        test_loader = DataLoader(
            dataset.torchDataset(test_index),
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=True,
            pin_memory=True,
        )
        test_loss, test_acc = self.evaluate(test_loader)
        LOG.info(f" [LOSS] TEST {test_loss}")
        LOG.info(f" [ACC] TEST {test_acc}")
        return test_loss, test_acc

    def _get_from_hub(self, repo=None):
        if repo is None:
            repo = "lamaroufle/meegnet"
        model_name = "_".join(self.name.split("_")[:-2])
        filename = f"{model_name}_{"_".join(map(str, self.input_size))}_{self.n_outputs}"
        model_path = hf_hub_download(repo_id="lamaroufle/meegnet", filename=filename + ".pt")
        hf_hub_download(repo_id="lamaroufle/meegnet", filename=filename + ".mat")
        return model_path

    def feature_extraction_from_pretrained(self, repo=None):
        model_path = self._get_from_hub(repo)
        net_state, _, _ = self._load_net(model_path)
        feat_state_dict = OrderedDict()
        for key, value in net_state["state_dict"].items():
            if key.startswith("feature"):
                feat_state_dict[".".join(key.split(".")[1:])] = value
        self.net.feature_extraction.load_state_dict(feat_state_dict)

    def from_pretrained(self, repo=None):
        model_path = self._get_from_hub(repo)
        self.load(model_path)
        
    def _load_net(self, model_path=None):
        if model_path is None:
            model_path = os.path.join(self.save_path, self.name + ".pt")
        if os.path.exists(model_path):
            LOG.info("=> loading checkpoint '{}'".format(model_path))
            checkpoint = torch.load(model_path)
            net_state = checkpoint["state_dict"]
            optimizer_state = checkpoint["optimizer"]
        mat_path = model_path[:-2] + "mat"
        if os.path.exists(mat_path):
            mat_data = loadmat(mat_path)
        else:
            LOG.warning(
                f"Error while loading checkpoint from {model_path}"
            )
        return net_state, optimizer_state, mat_data

    def load(self, model_path=None):
        net_state, optimizer_state, mat_data = self._load_net(model_path)
        self.net.load_state_dict(net_state)
        self.optimizer.load_state_dict(optimizer_state)
        self.results = mat_data

    def save(self, model_path=None):
        if model_path is None:
            model_path = os.path.join(self.save_path, self.name + ".pt")
        mat_path = model_path[:-2] + "mat"
        torch.save(self.checkpoint, model_path)
        savemat(mat_path, self.results)

    def compute_accuracy(self, y_pred, target):
        # Compute accuracy from 2 vectors of labels.
        correct = torch.eq(y_pred.max(1)[1], target).sum().type(torch.FloatTensor)
        return correct / len(target)



import os
import random
import logging
import pandas as pd
import numpy as np
import torch
from torch.utils.data import random_split
from scipy.stats import zscore
from meegnet.utils import strip_string

LOG = logging.getLogger("meegnet")


class _InfiniteSampler(torch.utils.data.Sampler):
    """
    Wraps another Sampler to yield an infinite stream.

    Parameters
    ----------
    sampler : torch.utils.data.Sampler
        The sampler to be wrapped.
    """

    def __init__(self, sampler):
        self.sampler = sampler

    def __iter__(self):
        while True:
            for batch in self.sampler:
                yield batch


class InfiniteDataLoader:
    """
    Creates an infinite data loader.

    Parameters
    ----------
    dataset : torch.utils.data.Dataset
        The dataset to be loaded.
    batch_size : int
        The size of the batches to be loaded.
    num_workers : int, optional
        The number of workers to use for loading the data, by default 0.
    pin_memory : bool, optional
        If True, the data loader will copy tensors into CUDA pinned memory before returning them. This can make data
        transfer faster, but requires more memory.
    weights : torch.Tensor, optional
        A 1D tensor assigning a weight to each sample in the dataset. If not provided, all samples are assumed to have
        the same weight.

    Returns
    -------
    InfiniteDataLoader
        An infinite data loader.
    """

    def __init__(self, dataset, batch_size, num_workers=0, pin_memory=False, weights=None):
        super().__init__()

        if weights is not None:
            sampler = torch.utils.data.WeightedRandomSampler(
                weights, replacement=True, num_samples=batch_size
            )
        else:
            sampler = torch.utils.data.RandomSampler(dataset, replacement=True)

        if weights is None:
            weights = torch.ones(len(dataset))

        batch_sampler = torch.utils.data.BatchSampler(
            sampler, batch_size=batch_size, drop_last=True
        )

        self._infinite_iterator = iter(
            torch.utils.data.DataLoader(
                dataset,
                num_workers=num_workers,
                batch_sampler=_InfiniteSampler(batch_sampler),
                pin_memory=pin_memory,
            )
        )

    def __iter__(self):
        while True:
            yield next(self._infinite_iterator)

    def __len__(self):
        raise ValueError


def _string_to_int(array):
    array = np.array(array)
    for i, element in enumerate(np.unique(array)):
        array[array == element] = i
    return array.astype(int)


class Dataset:
    """Creates a dataset

    Parameters
    ----------
    sfreq : int, optional
        The sampling frequency, by default 500.
    n_subjects : int, optional
        The number of subjects. Default value is None, which means all subjects are processed.
    zscore : bool, optional
        If True, z-scoring is applied to the data, by default True.
    n_samples : int, optional
        The number of samples to include, by default None.
    sensortype : str, optional
        The type of sensor to use, by default None.
    lso : bool, optional
        Leave subjects out. If False, within-subject splitting is used, by default False.
    random_state : int, optional
        The random state for reproducibility, by default 0.

    Attributes
    ----------
    sfreq : int
        The sampling frequency.
    n_subjects : int
        The number of subjects.
    n_samples : int
        The number of samples for each subject.
    data : torch.Tensor
        The data.
    labels : torch.Tensor
        The labels.
    groups : list
        The groups.
    subject_list : list
        The subject list.
    """

    def __init__(
        self,
        sfreq=500,
        n_subjects=None,
        zscore=True,
        n_samples=None,
        sensortype=None,
        lso=False,
        random_state=0,
    ):
        self.sfreq = sfreq
        self.n_subjects = n_subjects
        self.zscore = zscore
        self.n_samples = n_samples
        self.lso = lso
        self.random_state = random_state
        self.sensors = self._select_sensors(sensortype)

        self.data = []
        self.labels = []
        self.groups = []
        self.subject_list = []
        self.targets = self.labels
        self.data_path = None

    def _load_csv(self, data_path, csv_name="participants_info.csv"):
        csv_file = os.path.join(data_path, csv_name)
        participants_df = (
            pd.read_csv(csv_file, index_col=0)
            .sample(frac=1, random_state=self.random_state)
            .reset_index(drop=True)[: self.n_subjects]
        )
        return participants_df

    def preload(self, data_path, csv_path=None):
        """loads the subject list from a csv file (participants_info.csv in the data_folder by default).

        Parameters
        ----------
        data_path :
            _description_
        csv_path : _type_, optional
            _description_, by default None

        Returns
        -------
        _type_
            _description_
        """

        if csv_path is not None and os.path.exists(csv_path):
            dataframe = pd.read_csv(csv_path, index_col=0)
        else:
            dataframe = self._load_csv(data_path)
        LOG.info(f"Logging subjects and labels from {data_path}...")
        self.data_path = data_path
        self.subject_list = list(dataframe["sub"])
        return dataframe

    def load(self, data_path=None, csv_path=None, one_sub=None):
        """Loads the data from the "downsamples_[sfreq]" folder in the data_path.

        Parameters
        ----------
        data_path : _type_, optional
            _description_, by default None
        csv_path : _type_, optional
            _description_, by default None
        one_sub : _type_, optional
            _description_, by default None
        """
        assert self.data_path is not None or data_path is not None, "data_path must be set."
        if self.data_path is None:
            self.data_path = data_path
        dataframe = self.preload(self.data_path, csv_path)
        if one_sub is None:
            LOG.info(f"Found {len(self.subject_list)} subjects to load.")
        else:
            LOG.info(f"Loading subject {one_sub}")
        data_folder = f"downsampled_{self.sfreq}"
        numpy_filepath = os.path.join(self.data_path, data_folder)
        for file in os.listdir(numpy_filepath):
            np.random.seed(self.random_state)
            random.seed(self.random_state)
            torch.manual_seed(self.random_state)
            sub = file.split("_")[0]
            if one_sub is not None:
                if sub != one_sub:
                    continue
            if sub in self.subject_list:
                row = dataframe.loc[dataframe["sub"] == sub]
                sub_data = self._load_sub(os.path.join(numpy_filepath, file))
                if sub_data is None:
                    continue
                if self.sensors is not None:
                    sub_data = sub_data[:, self.sensors, :, :]
                labels = list(map(strip_string, row["label"].item().split(", ")))
                if len(labels) == 1:
                    if labels[0] in self.subject_list:
                        labels = [self.subject_list.index(labels[0])] * len(sub_data)
                    else:
                        labels = [labels[0]] * len(sub_data)
                if len(labels) != len(sub_data):
                    LOG.warning(
                        "Length of label vector different from number "
                        f"of data samples for subject {sub}. Skipping."
                    )
                    continue
                if self.n_samples is not None:
                    if self.n_samples > len(sub_data):
                        LOG.warning(
                            f"Number of available samples for {file} "
                            f"below the requested amount ({self.n_samples})",
                        )
                    random_samples = np.random.choice(
                        np.arange(len(sub_data)),
                        self.n_samples,
                        replace=False,
                    )
                    sub_data = sub_data[random_samples]
                    labels = np.array(labels)[random_samples]

                self.data.append(torch.Tensor(sub_data))
                self.labels.append(np.array(labels))
                self.groups += [sub] * len(labels)

        if len(self) == 0:
            return
        elif len(self) == 1:
            self.data = self.data[0]
            self.labels = self.labels[0]
        else:
            self.data = torch.cat(self.data, 0)
            self.labels = np.concatenate(self.labels, axis=0)

        if len(self.data.shape) != 4:
            self.data = self.data[:, np.newaxis]

        self.set_labels(self.labels)

        if type(self.groups[0]) != int:
            self.groups = _string_to_int(self.groups)
        self.groups = torch.Tensor(self.groups)

        self.n_subjects = len(np.unique(self.groups))

    def __len__(self):
        """Returns the length of the dataset (total number of data examples).

        Returns
        -------
        _type_
            _description_
        """
        return len(self.data)

    def _load_sub(self, filepath):
        """Loads a subject's data.

        Parameters
        ----------
        filepath : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        try:
            data = np.load(filepath)
        except IOError:
            LOG.warning(f"There was a problem loading subject {filepath}")
            return None
        if self.zscore:
            data = np.array(list(map(lambda x: zscore(x, axis=-1), data)))
        return torch.Tensor(np.array(data))

    def _select_sensors(self, sensortype):
        """For MEG data only. Selects the sensors slices from the data.

        Parameters
        ----------
        sensortype : str
            type of sensor. Must be either "MAG", "plannar1", "plannar2" or "plannar".

        Returns
        -------
        _type_
            _description_
        """
        if sensortype == "MAG":
            return [0]
        elif sensortype in ["GRAD", "plannar"]:
            return [1, 2]
        elif sensortype in ["GRAD1", "plannar1"]:
            return [1]
        elif sensortype in ["GRAD2", "plannar2"]:
            return [2]
        else:
            return None

    def _assert_sizes(self, train_size, valid_size, test_size=None):
        """Asserts that the sum of the data ratios is equal to 1.

        Parameters
        ----------
        train_size : _type_
            _description_
        valid_size : _type_
            _description_
        test_size : _type_, optional
            _description_, by default None
        """
        if test_size is None:
            test_size = 0
        assert (
            sum((train_size, valid_size, test_size)) == 1
        ), "sum of data ratios must be equal to 1"

    def _within_subject_split(self, sizes, generator):
        """Splits the data within each subject using the specified sizes and generator.

        Parameters
        ----------
        sizes : _type_
            _description_
        generator : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        indexes = []
        index_groups = [[] for _ in range(self.n_subjects)]
        for index, group in enumerate(self.groups):
            index_groups[group].append(index)
        for group in index_groups:
            indexes.append(random_split(group, sizes, generator))
        # logging.info(
        #     f"Using {self.n_subjects} subjects: {train_size} for train, "
        #     f"{valid_size} for validation, and {test_size} for test"
        # )
        return (sum([list(index[i]) for index in indexes], []) for i in range(3))

    def data_split(self, train_size, valid_size, test_size=None):
        """Splits the data into training, validation, and test sets based on the specified sizes and stratification (if desired).

        Parameters
        ----------
        train_size : _type_
            _description_
        valid_size : _type_
            _description_
        test_size : _type_, optional
            _description_, by default None

        Returns
        -------
        _type_
            _description_
        """
        # TODO add stratification for the data splits
        self._assert_sizes(train_size, valid_size, test_size)
        generator = torch.Generator().manual_seed(self.random_state)
        if self.lso:
            return self._within_subject_split((train_size, valid_size, test_size), generator)
        else:
            return random_split(
                np.arange(len(self)), [train_size, valid_size, test_size], generator
            )

    def torchDataset(self, index):
        """Returns a Torch dataset instance of the torch Dataset class for the given index.

        Parameters
        ----------
        index : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        return torch.utils.data.TensorDataset(self.data[index], self.labels[index])

    def set_labels(self, labels):
        if type(labels[0]) in (str, np.str_):
            labels = _string_to_int(labels)
        self.labels = torch.Tensor(labels, dtype=torch.int32)


class RestDataset(Dataset):
    """
    Creates a dataset for deep learning models from REST data with windowing.

    Parameters
    ----------
    window : int, optional
        The window size in seconds, by default 2.
    overlap : float, optional
        The overlap between windows, by default 0.
    offset : int, optional
        The offset in seconds, by default 10.
    sfreq : int, optional
        The sampling frequency, by default 500.
    n_subjects : int, optional
        The number of subjects. Default value is None, which means all subjects are processed.
    zscore : bool, optional
        If True, z-scoring is applied to the data, by default True.
    n_samples : int, optional
        The number of samples to include, by default None.
    sensortype : str, optional
        The type of sensor to use, by default None.
    lso : bool, optional
        Leave subjects out. If False, within-subject splitting is used, by default False.
    random_state : int, optional
        The random state for reproducibility, by default 0.

    Attributes
    ----------
    window : int
        The window size in seconds.
    overlap : float
        The overlap between windows.
    offset : int
        The offset in seconds.
    sfreq : int
        The sampling frequency.
    n_subjects : int
        The number of subjects.
    n_samples : int
        The number of samples for each subject.
    data : torch.Tensor
        The data.
    labels : torch.Tensor
        The labels.
    groups : list
        The groups.
    subject_list : list
        The subject list.

    Methods
    -------
    _load_sub(filepath)
        Loads a subject's data with windowing and overlap.
    """

    def __init__(
        self,
        window=2,
        overlap=0,
        offset=10,
        sfreq=500,
        n_subjects=None,
        zscore=True,
        n_samples=None,
        sensortype=None,
        lso=False,
        random_state=0,
    ):

        Dataset.__init__(
            self, sfreq, n_subjects, zscore, n_samples, sensortype, lso, random_state
        )

        assert 0 <= overlap < 1, "Overlap must be between 0 and 1."
        self.window = window
        self.overlap = overlap
        self.offset = offset

    def _load_sub(self, filepath):
        try:
            step = int(self.window * self.sfreq * (1 - self.overlap))
            start = int(self.offset * self.sfreq)
            sub_data = np.load(filepath)
            if len(sub_data.shape) < 3:
                sub_data = sub_data[np.newaxis, :]
            data = []
            for i in range(start, sub_data.shape[-1], step):
                trial = sub_data[:, :, i : i + step]
                if trial.shape[-1] == step:
                    if not np.isnan(trial).any():
                        if self.zscore:
                            trial = zscore(trial, axis=-1)
                        data.append(trial)
        except IOError:
            LOG.warning(f"There was a problem loading subject {filepath}")
            return None
        except ValueError:
            LOG.warning(f"There was a problem loading subject {filepath}")
            return None
        return torch.Tensor(np.array(data))



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models
from domainbed.lib import wide_resnet
import copy


def remove_batch_norm_from_resnet(model):
    fuse = torch.nn.utils.fusion.fuse_conv_bn_eval
    model.eval()

    model.conv1 = fuse(model.conv1, model.bn1)
    model.bn1 = Identity()

    for name, module in model.named_modules():
        if name.startswith("layer") and len(name) == 6:
            for b, bottleneck in enumerate(module):
                for name2, module2 in bottleneck.named_modules():
                    if name2.startswith("conv"):
                        bn_name = "bn" + name2[-1]
                        setattr(bottleneck, name2,
                                fuse(module2, getattr(bottleneck, bn_name)))
                        setattr(bottleneck, bn_name, Identity())
                if isinstance(bottleneck.downsample, torch.nn.Sequential):
                    bottleneck.downsample[0] = fuse(bottleneck.downsample[0],
                                                    bottleneck.downsample[1])
                    bottleneck.downsample[1] = Identity()
    model.train()
    return model


class Identity(nn.Module):
    """An identity layer"""
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


class MLP(nn.Module):
    """Just  an MLP"""
    def __init__(self, n_inputs, n_outputs, hparams):
        super(MLP, self).__init__()
        self.input = nn.Linear(n_inputs, hparams['mlp_width'])
        self.dropout = nn.Dropout(hparams['mlp_dropout'])
        self.hiddens = nn.ModuleList([
            nn.Linear(hparams['mlp_width'], hparams['mlp_width'])
            for _ in range(hparams['mlp_depth']-2)])
        self.output = nn.Linear(hparams['mlp_width'], n_outputs)
        self.n_outputs = n_outputs

    def forward(self, x):
        x = self.input(x)
        x = self.dropout(x)
        x = F.relu(x)
        for hidden in self.hiddens:
            x = hidden(x)
            x = self.dropout(x)
            x = F.relu(x)
        x = self.output(x)
        return x


class ResNet(torch.nn.Module):
    """ResNet with the softmax chopped off and the batchnorm frozen"""
    def __init__(self, input_shape, hparams):
        super(ResNet, self).__init__()
        if hparams['resnet18']:
            self.network = torchvision.models.resnet18(pretrained=True)
            self.n_outputs = 512
        else:
            self.network = torchvision.models.resnet50(pretrained=True)
            self.n_outputs = 2048

        # self.network = remove_batch_norm_from_resnet(self.network)

        # adapt number of channels
        nc = input_shape[0]
        if nc != 3:
            tmp = self.network.conv1.weight.data.clone()

            self.network.conv1 = nn.Conv2d(
                nc, 64, kernel_size=(7, 7),
                stride=(2, 2), padding=(3, 3), bias=False)

            for i in range(nc):
                self.network.conv1.weight.data[:, i, :, :] = tmp[:, i % 3, :, :]

        # save memory
        del self.network.fc
        self.network.fc = Identity()

        self.freeze_bn()
        self.hparams = hparams
        self.dropout = nn.Dropout(hparams['resnet_dropout'])

    def forward(self, x):
        """Encode x into a feature vector of size n_outputs."""
        return self.dropout(self.network(x))

    def train(self, mode=True):
        """
        Override the default train() to freeze the BN parameters
        """
        super().train(mode)
        self.freeze_bn()

    def freeze_bn(self):
        for m in self.network.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()


class MNIST_CNN(nn.Module):
    """
    Hand-tuned architecture for MNIST.
    Weirdness I've noticed so far with this architecture:
    - adding a linear layer after the mean-pool in features hurts
        RotatedMNIST-100 generalization severely.
    """
    n_outputs = 128

    def __init__(self, input_shape):
        super(MNIST_CNN, self).__init__()
        self.conv1 = nn.Conv2d(input_shape[0], 64, 3, 1, padding=1)
        self.conv2 = nn.Conv2d(64, 128, 3, stride=2, padding=1)
        self.conv3 = nn.Conv2d(128, 128, 3, 1, padding=1)
        self.conv4 = nn.Conv2d(128, 128, 3, 1, padding=1)

        self.bn0 = nn.GroupNorm(8, 64)
        self.bn1 = nn.GroupNorm(8, 128)
        self.bn2 = nn.GroupNorm(8, 128)
        self.bn3 = nn.GroupNorm(8, 128)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.bn0(x)

        x = self.conv2(x)
        x = F.relu(x)
        x = self.bn1(x)

        x = self.conv3(x)
        x = F.relu(x)
        x = self.bn2(x)

        x = self.conv4(x)
        x = F.relu(x)
        x = self.bn3(x)

        x = self.avgpool(x)
        x = x.view(len(x), -1)
        return x


class ContextNet(nn.Module):
    def __init__(self, input_shape):
        super(ContextNet, self).__init__()

        # Keep same dimensions
        padding = (5 - 1) // 2
        self.context_net = nn.Sequential(
            nn.Conv2d(input_shape[0], 64, 5, padding=padding),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Conv2d(64, 64, 5, padding=padding),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Conv2d(64, 1, 5, padding=padding),
        )

    def forward(self, x):
        return self.context_net(x)


def Featurizer(input_shape, hparams):
    """Auto-select an appropriate featurizer for the given input shape."""
    if len(input_shape) == 1:
        return MLP(input_shape[0], hparams["mlp_width"], hparams)
    elif input_shape[1:3] == (28, 28):
        return MNIST_CNN(input_shape)
    elif input_shape[1:3] == (32, 32):
        return wide_resnet.Wide_ResNet(input_shape, 16, 2, 0.)
    elif input_shape[1:3] == (224, 224):
        return ResNet(input_shape, hparams)
    else:
        raise NotImplementedError


def Classifier(in_features, out_features, is_nonlinear=False):
    if is_nonlinear:
        return torch.nn.Sequential(
            torch.nn.Linear(in_features, in_features // 2),
            torch.nn.ReLU(),
            torch.nn.Linear(in_features // 2, in_features // 4),
            torch.nn.ReLU(),
            torch.nn.Linear(in_features // 4, out_features))
    else:
        return torch.nn.Linear(in_features, out_features)


class WholeFish(nn.Module):
    def __init__(self, input_shape, num_classes, hparams, weights=None):
        super(WholeFish, self).__init__()
        featurizer = Featurizer(input_shape, hparams)
        classifier = Classifier(
            featurizer.n_outputs,
            num_classes,
            hparams['nonlinear_classifier'])
        self.net = nn.Sequential(
            featurizer, classifier
        )
        if weights is not None:
            self.load_state_dict(copy.deepcopy(weights))

    def reset_weights(self, weights):
        self.load_state_dict(copy.deepcopy(weights))

    def forward(self, x):
        return self.net(x)



import configparser
import configargparse

parser = configargparse.ArgParser()


def save_config(args: dict, config_filepath: str = "config.ini"):
    """Saves a dictionnary to a given path.

    Args:
        args (dict): The argument dictionnary to be saved to the config filepath.
        config_filepath (str): The filepath to be used to save the config file.
    """
    with open(config_filepath, "w") as conf:
        config_object = configparser.ConfigParser()
        config_object.add_section("config")
        for key, value in args.items():
            config_object.set("config", key.replace("_", "-"), str(value))
        config_object.write(conf)
    return


parser.add(
    "-c",
    "--config",
    is_config_file=True,
    default="../default.ini",
    help="config file path",
)
parser.add(
    "--testsplit",
    type=int,
    default=-1,
    help="Will remove the 20% holdout set by default and usit for cross-val. Using 5-Fold instead of 4-Fold.",
)
parser.add(
    "--flat",
    action="store_true",
    help="will flatten sensors dimension",
)
parser.add(
    "--epoched",
    action="store_true",
    help="will load epoched data instead of the whole signal for each subject. This option only works for passive and smt data in which there are events that we use to generate epochs.",
)
parser.add(
    "--randomsearch",
    action="store_true",
    help="Launches one cross-val on a subset of data or full random search depending on testsplit parameter",
)
parser.add(
    "--fold",
    default=-1,
    type=int,
    help="will only do a specific fold if specified. must be between 0 and 3, or 0 and 4 if testsplit option is true",
)
parser.add(
    "--net-option",
    default="best_net",
    choices=["custom_net", "best_net", "EEGNet", "VGG", "vanPutNet", "MLP"],
    help="cNet is the custom net.",
)
parser.add(
    "--datatype",
    default="rest",
    choices=["rest", "smt", "passive"],
    help="the type of data to be loaded",
)
parser.add(
    "--crossval",
    action="store_true",
    help="wether to do a 4-FOLD cross-validation on the train+valid set.",
)
parser.add(
    "--n-samples",
    type=int,
    default=-1,
    help="limit of number of samples per subjects",
)
parser.add(
    "-f", "--filters", default=8, type=int, help="The size of the first convolution"
)
parser.add(
    "--segment-length",
    type=float,
    default=2,
    help="The length (in seconds) of the segment to consider for resting-state data.",
)
parser.add(
    "--sfreq",
    type=int,
    default=500,
    help="The sampling frequency of the data, mainly for loading the correct data",
)
parser.add(
    "--maxpool",
    type=int,
    default=0,
    help="adds a maxpool layer in between conv layers",
)
parser.add(
    "--batchnorm",
    action="store_true",
    help="adds a batchnorm layer in between conv layers",
)
parser.add(
    "--permute-labels",
    action="store_true",
    help="Permutes the labesl in order to test for chance level",
)
parser.add(
    "--printmem",
    action="store_true",
    help="Shows RAM information before and during the data loading process.",
)
parser.add(
    "--log",
    action="store_true",
    help="stores all prints in a logfile instead of printing them",
)
parser.add(
    "--lr",
    type=float,
    default=0.00001,
    help="the starting learning rate of the optimizer",
)
parser.add(
    "--hlayers",
    type=int,
    default=1,
    help="number of hidden layers",
)
parser.add(
    "--patience",
    type=int,
    default=20,
    help="patience for early stopping",
)
parser.add(
    "--model-name",
    type=str,
    default="net",
    help="Name of the network for file_save",
)
parser.add(
    "--num-workers",
    type=int,
    default=4,
    help="number of workers to load data while gpu is processing",
)
parser.add(
    "--train-size",
    type=float,
    default=0.8,
    help="The proportion of data to use in the train set",
)
parser.add(
    "--model-path",
    type=str,
    default=None,
    help="The default path to save all computed data, model and visualisations.",
)
parser.add(
    "--save-path",
    type=str,
    default=".",
    help="The default path to save all computed data, model and visualisations.",
)
parser.add(
    "--raw-path",
    type=str,
    default=None,
    help="The path where the raw data can be found.",
)
parser.add(
    "--visu-path",
    type=str,
    default=None,
    help="The path where the visualisation matrices wil be saved to and loaded from.",
)
parser.add(
    "--processed-path",
    type=str,
    default=None,
    help="The path where the data samples can be found.",
)
parser.add(
    "--seed",
    default=42,
    type=int,
    help="Seed to use for random splits.",
)
parser.add(
    "--max-subj",
    default=1000,
    type=int,
    help="maximum number of subjects to use (1000 uses all subjects)",
)
parser.add(
    "--clf-type",
    default="",
    choices=["eventclf", "subclf", "toneclf"],
    help="The type of classification to run.",
)
parser.add(
    "-s",
    "--sensors",
    default="MAG",
    choices=["GRAD", "MAG", "ALL"],
    help="The type of sensors to keep, default=MAG",
)
parser.add(
    "--feature",
    default="temporal",
    choices=["temporal", "bands", "bins"],
    help="Data type to use.",
)
parser.add(
    "-b",
    "--batch-size",
    default=128,
    type=int,
    help="The batch size used for learning.",
)
parser.add(
    "-d",
    "--dropout",
    default=0.25,
    type=float,
    help="The dropout rate of the linear layers",
)
parser.add(
    "--times",
    action="store_true",
    help="Instead of running the training etc, run a series of test in order to choose best set of workers and batch sizes to get faster epochs.",
)
parser.add(
    "--chunkload",
    action="store_true",
    help="Chunks the data and loads data batch per batch. Will be slower but is necessary when RAM size is too low to handle whole dataset.",
)
parser.add(
    "--debug",
    action="store_true",
    help="loads dummy data in the net to ensure everything is working fine",
)
parser.add(
    "--dropout-option",
    default="same",
    choices=["same", "double", "inverted"],
    help="sets if the first dropout and the second are the same or if the first one or the second one should be bigger",
)
parser.add(
    "-l", "--linear", default=100, type=int, help="The size of the second linear layer"
)
parser.add(
    "-n",
    "--nchan",
    default=100,
    type=int,
    help="the number of channels for the first convolution, the other channel numbers scale with this one.",
)
parser.add(
    "--compute-psd",
    action="store_true",
    help="wether or not to to compute psd using saliency windows in the compute_saliency_maps.py script.",
)
parser.add(
    "--w-size",
    type=int,
    default=300,
    help="The window size for saliency based psd computation in the compute_saliency_maps.py script.",
)
parser.add(
    "--overlap",
    type=float,
    default=0,
    help="the overlap value between segments for continous data.",
)
parser.add(
    "--confidence",
    type=float,
    default=0.98,
    help="the confidence threshold needed for a trial to be selected for visualisation in the compute_ and visu_saliency_maps.py script.",
)



import os
import mne
from mne.viz import plot_topomap
import copy
import torch
from torch.autograd import Variable
from torchvision import models
import numpy as np
from PIL import Image
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.pyplot as plt
import matplotlib.cm as mpl_color_map
from matplotlib.gridspec import GridSpec
from collections.abc import Iterable
from joblib import Parallel, delayed
from meegnet.utils import compute_psd


def convert_to_grayscale(im_as_arr):
    """
        Converts 3d image to grayscale

    Args:
        im_as_arr (numpy arr): RGB image with shape (D,W,H)

    returns:
        grayscale_im (numpy_arr): Grayscale image with shape (1,W,D)
    """
    grayscale_im = np.sum(np.abs(im_as_arr), axis=0)
    im_max = np.percentile(grayscale_im, 99)
    im_min = np.min(grayscale_im)
    grayscale_im = np.clip((grayscale_im - im_min) / (im_max - im_min), 0, 1)
    grayscale_im = np.expand_dims(grayscale_im, axis=0)
    return grayscale_im


def save_gradient_images(gradient, file_name):
    """
        Exports the original gradient image

    Args:
        gradient (np arr): Numpy array of the gradient with shape (3, 224, 224)
        file_name (str): File name to be exported
    """
    if not os.path.exists("../results"):
        os.makedirs("../results")
    # Normalize
    gradient = gradient - gradient.min()
    gradient /= gradient.max()
    # Save image
    # path_to_file = os.path.join("../results", file_name + ".jpg")
    path_to_file = os.path.join(file_name + ".jpg")
    save_image(gradient, path_to_file)


def save_class_activation_images(org_img, activation_map, file_name):
    """
        Saves cam activation map and activation map on the original image

    Args:
        org_img (PIL img): Original image
        activation_map (numpy arr): Activation map (grayscale) 0-255
        file_name (str): File name of the exported image
    """
    if not os.path.exists("../results"):
        os.makedirs("../results")
    # Grayscale activation map
    heatmap, heatmap_on_image = apply_colormap_on_image(org_img, activation_map, "hsv")
    # Save colored heatmap
    path_to_file = os.path.join("../results", file_name + "_Cam_Heatmap.png")
    save_image(heatmap, path_to_file)
    # Save heatmap on iamge
    path_to_file = os.path.join("../results", file_name + "_Cam_On_Image.png")
    save_image(heatmap_on_image, path_to_file)
    # SAve grayscale heatmap
    path_to_file = os.path.join("../results", file_name + "_Cam_Grayscale.png")
    save_image(activation_map, path_to_file)


def apply_colormap_on_image(org_im, activation, colormap_name):
    """
        Apply heatmap on image
    Args:
        org_img (PIL img): Original image
        activation_map (numpy arr): Activation map (grayscale) 0-255
        colormap_name (str): Name of the colormap
    """
    # Get colormap
    color_map = mpl_color_map.get_cmap(colormap_name)
    no_trans_heatmap = color_map(activation)
    # Change alpha channel in colormap to make sure original image is displayed
    heatmap = copy.copy(no_trans_heatmap)
    heatmap[:, :, 3] = 0.4
    heatmap = Image.fromarray((heatmap * 255).astype(np.uint8))
    no_trans_heatmap = Image.fromarray((no_trans_heatmap * 255).astype(np.uint8))

    # Apply heatmap on iamge
    heatmap_on_image = Image.new("RGBA", org_im.size)
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, org_im.convert("RGBA"))
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, heatmap)
    return no_trans_heatmap, heatmap_on_image


def format_np_output(np_arr):
    """
        This is a (kind of) bandaid fix to streamline saving procedure.
        It converts all the outputs to the same format which is 3xWxH
        with using sucecssive if clauses.
    Args:
        im_as_arr (Numpy array): Matrix of shape 1xWxH or WxH or 3xWxH
    """
    # Phase/Case 1: The np arr only has 2 dimensions
    # Result: Add a dimension at the beginning
    if len(np_arr.shape) == 2:
        np_arr = np.expand_dims(np_arr, axis=0)
    # Phase/Case 2: Np arr has only 1 channel (assuming first dim is channel)
    # Result: Repeat first channel and convert 1xWxH to 3xWxH
    if np_arr.shape[0] == 1:
        np_arr = np.repeat(np_arr, 3, axis=0)
    # Phase/Case 3: Np arr is of shape 3xWxH
    # Result: Convert it to WxHx3 in order to make it saveable by PIL
    if np_arr.shape[0] == 3:
        np_arr = np_arr.transpose(1, 2, 0)
    # Phase/Case 4: NP arr is normalized between 0-1
    # Result: Multiply with 255 and change type to make it saveable by PIL
    if np.max(np_arr) <= 1:
        np_arr = (np_arr * 255).astype(np.uint8)
    return np_arr


def save_image(im, path):
    """
        Saves a numpy matrix or PIL image as an image
    Args:
        im_as_arr (Numpy array): Matrix of shape DxWxH
        path (str): Path to the image
    """
    if isinstance(im, (np.ndarray, np.generic)):
        im = format_np_output(im)
        im = Image.fromarray(im)
    im.save(path)


def preprocess_image(pil_im, resize_im=True):
    """
        Processes image for CNNs

    Args:
        PIL_img (PIL_img): PIL Image or numpy array to process
        resize_im (bool): Resize to 224 or not
    returns:
        im_as_var (torch variable): Variable that contains processed float tensor
    """
    # mean and std list for channels (Imagenet)
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]

    # ensure or transform incoming image to PIL image
    if type(pil_im) != Image.Image:
        try:
            pil_im = Image.fromarray(pil_im)
        except Exception as e:
            print("could not transform PIL_img to a PIL Image object. Please check input.")

    # Resize image
    if resize_im:
        pil_im = pil_im.resize((224, 224), Image.ANTIALIAS)

    im_as_arr = np.float32(pil_im)
    im_as_arr = im_as_arr.transpose(2, 0, 1)  # Convert array to D,W,H
    # Normalize the channels
    for channel, _ in enumerate(im_as_arr):
        im_as_arr[channel] /= 255
        im_as_arr[channel] -= mean[channel]
        im_as_arr[channel] /= std[channel]
    # Convert to float tensor
    im_as_ten = torch.from_numpy(im_as_arr).float()
    # Add one more channel to the beginning. Tensor shape = 1,3,224,224
    im_as_ten.unsqueeze_(0)
    # Convert to Pytorch variable
    im_as_var = Variable(im_as_ten, requires_grad=True)
    return im_as_var


def recreate_image(im_as_var):
    """
        Recreates images from a torch variable, sort of reverse preprocessing
    Args:
        im_as_var (torch variable): Image to recreate
    returns:
        recreated_im (numpy arr): Recreated image in array
    """
    reverse_mean = [-0.485, -0.456, -0.406]
    reverse_std = [1 / 0.229, 1 / 0.224, 1 / 0.225]
    recreated_im = copy.copy(im_as_var.data.numpy()[0])
    for c in range(3):
        recreated_im[c] /= reverse_std[c]
        recreated_im[c] -= reverse_mean[c]
    recreated_im[recreated_im > 1] = 1
    recreated_im[recreated_im < 0] = 0
    recreated_im = np.round(recreated_im * 255)

    recreated_im = np.uint8(recreated_im).transpose(1, 2, 0)
    return recreated_im


def choose_best_window(data, fs=500, w_size=300):
    """
    data: array
        Must be of size k x n_samples. k can be sensor dimension or trial dimension.
    w_size: int
        The size of the window in ms
    """
    # Generate masks over time samples where sample is superior to 2x std of the vector
    masks = [dat >= (np.mean(dat) + np.std(dat) * 2) for dat in data]
    w_size = int(w_size * fs / 1000)
    best_window_idx = Parallel(n_jobs=-1)(delayed(best_window)(mask, w_size) for mask in masks)
    return best_window_idx


def best_window(mask, w_size):
    # create an array of all the possible windows in the trial with their mask values
    windows = np.array([mask[i : i + w_size] for i in range(len(mask) - w_size)])
    # get the sum of mask values in each windows. High number indicates lots of True values
    values = [sum(window) for window in windows]
    if max(values) == 0:
        return None
    # best window is where the values are the highest
    best_window_index = np.where(values == max(values))[0]
    # below is to handle if there are multiple best windows with the same values, in which case we
    # try to check if they overlap, and keep only if they are separated.
    if len(best_window_index) > 1:
        idx_range = best_window_index[-1] - best_window_index[0]
        # duration betweeen first and last window must be w_size/2
        if idx_range <= int(w_size / 2):
            # Then best would be in the middle of all this
            best = int(len(best_window_index) / 2)
            return best_window_index[best]
        elif idx_range <= w_size:
            return (best_window_index[0], best_window_index[-1])
        else:
            best = int(len(best_window_index) / 2)
            return (best_window_index[0], best, best_window_index[-1])
    else:
        # only one window, just add it.
        return best_window_index[0]


def compute_single_sal_psd(index, trial, w_size, fs):
    if isinstance(index, Iterable):
        tmp = []
        for idx in index:
            window = trial[idx : idx + w_size]
            while len(window.shape) < 3:
                window = window[np.newaxis, :]
            tmp.append(compute_psd(window, fs=fs))
        return np.mean(tmp, axis=0)
    else:
        if index is not None:
            window = trial[index : index + w_size]
            while len(window.shape) < 3:
                window = window[np.newaxis, :]
            return compute_psd(window, fs=fs)
        else:
            return [None] * 7


def compute_saliency_based_psd(saliency, trial, w_size, fs):
    chan_data = []
    for i, chan in enumerate(saliency):
        windows_idx = choose_best_window(chan, fs, w_size)
        transformed_data = Parallel(n_jobs=1)(
            delayed(compute_single_sal_psd)(index, trial[i, j], w_size, fs)
            for j, index in enumerate(windows_idx)
        )
        chan_data.append(np.array(transformed_data).squeeze())
    return np.array(chan_data)


def get_positive_negative_saliency(gradient):
    """
        Generates positive and negative saliency maps based on the gradient
    Args:
        gradient (numpy arr): Gradient of the operation to visualize

    returns:
        pos_saliency ( )
    """
    pos_saliency = np.maximum(0, gradient) / gradient.max()
    neg_saliency = np.maximum(0, -gradient) / -gradient.min()
    return pos_saliency, neg_saliency


def get_example_params(example_index):
    """
        Gets used variables for almost all visualizations, like the image, model etc.

    Args:
        example_index (int): Image id to use from examples

    returns:
        original_image (numpy arr): Original image read from the file
        prep_img (numpy_arr): Processed image
        target_class (int): Target class for the image
        file_name_to_export (string): File name to export the visualizations
        pretrained_model(Pytorch model): Model to use for the operations
    """
    # Pick one of the examples
    example_list = (
        ("../input_images/snake.jpg", 56),
        ("../input_images/cat_dog.png", 243),
        ("../input_images/spider.png", 72),
    )
    img_path = example_list[example_index][0]
    target_class = example_list[example_index][1]
    file_name_to_export = img_path[img_path.rfind("/") + 1 : img_path.rfind(".")]
    # Read image
    original_image = Image.open(img_path).convert("RGB")
    # Process image
    prep_img = preprocess_image(original_image)
    # Define model
    pretrained_model = models.alexnet(pretrained=True)
    return (
        original_image,
        prep_img,
        target_class,
        file_name_to_export,
        pretrained_model,
    )


def convert_to_grayscale(im_as_arr):
    """
        Converts 3d image to grayscale

    Args:
        im_as_arr (numpy arr): RGB image with shape (D,W,H)

    returns:
        grayscale_im (numpy_arr): Grayscale image with shape (1,W,D)
    """
    grayscale_im = np.sum(np.abs(im_as_arr), axis=0)
    im_max = np.percentile(grayscale_im, 99)
    im_min = np.min(grayscale_im)
    grayscale_im = np.clip((grayscale_im - im_min) / (im_max - im_min), 0, 1)
    grayscale_im = np.expand_dims(grayscale_im, axis=0)
    return grayscale_im


def save_gradient_images(gradient, file_name):
    """
        Exports the original gradient image

    Args:
        gradient (np arr): Numpy array of the gradient with shape (3, 224, 224)
        file_name (str): File name to be exported
    """
    if not os.path.exists("../results"):
        os.makedirs("../results")
    # Normalize
    gradient = gradient - gradient.min()
    gradient /= gradient.max()
    # Save image
    # path_to_file = os.path.join("../results", file_name + ".jpg")
    path_to_file = os.path.join(file_name + ".jpg")
    save_image(gradient, path_to_file)


def save_class_activation_images(org_img, activation_map, file_name):
    """
        Saves cam activation map and activation map on the original image

    Args:
        org_img (PIL img): Original image
        activation_map (numpy arr): Activation map (grayscale) 0-255
        file_name (str): File name of the exported image
    """
    if not os.path.exists("../results"):
        os.makedirs("../results")
    # Grayscale activation map
    heatmap, heatmap_on_image = apply_colormap_on_image(org_img, activation_map, "hsv")
    # Save colored heatmap
    path_to_file = os.path.join("../results", file_name + "_Cam_Heatmap.png")
    save_image(heatmap, path_to_file)
    # Save heatmap on iamge
    path_to_file = os.path.join("../results", file_name + "_Cam_On_Image.png")
    save_image(heatmap_on_image, path_to_file)
    # SAve grayscale heatmap
    path_to_file = os.path.join("../results", file_name + "_Cam_Grayscale.png")
    save_image(activation_map, path_to_file)


def apply_colormap_on_image(org_im, activation, colormap_name):
    """
        Apply heatmap on image
    Args:
        org_img (PIL img): Original image
        activation_map (numpy arr): Activation map (grayscale) 0-255
        colormap_name (str): Name of the colormap
    """
    # Get colormap
    color_map = mpl_color_map.get_cmap(colormap_name)
    no_trans_heatmap = color_map(activation)
    # Change alpha channel in colormap to make sure original image is displayed
    heatmap = copy.copy(no_trans_heatmap)
    heatmap[:, :, 3] = 0.4
    heatmap = Image.fromarray((heatmap * 255).astype(np.uint8))
    no_trans_heatmap = Image.fromarray((no_trans_heatmap * 255).astype(np.uint8))

    # Apply heatmap on iamge
    heatmap_on_image = Image.new("RGBA", org_im.size)
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, org_im.convert("RGBA"))
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, heatmap)
    return no_trans_heatmap, heatmap_on_image


def format_np_output(np_arr):
    """
        This is a (kind of) bandaid fix to streamline saving procedure.
        It converts all the outputs to the same format which is 3xWxH
        with using sucecssive if clauses.
    Args:
        im_as_arr (Numpy array): Matrix of shape 1xWxH or WxH or 3xWxH
    """
    # Phase/Case 1: The np arr only has 2 dimensions
    # Result: Add a dimension at the beginning
    if len(np_arr.shape) == 2:
        np_arr = np.expand_dims(np_arr, axis=0)
    # Phase/Case 2: Np arr has only 1 channel (assuming first dim is channel)
    # Result: Repeat first channel and convert 1xWxH to 3xWxH
    if np_arr.shape[0] == 1:
        np_arr = np.repeat(np_arr, 3, axis=0)
    # Phase/Case 3: Np arr is of shape 3xWxH
    # Result: Convert it to WxHx3 in order to make it saveable by PIL
    if np_arr.shape[0] == 3:
        np_arr = np_arr.transpose(1, 2, 0)
    # Phase/Case 4: NP arr is normalized between 0-1
    # Result: Multiply with 255 and change type to make it saveable by PIL
    if np.max(np_arr) <= 1:
        np_arr = (np_arr * 255).astype(np.uint8)
    return np_arr


def save_image(im, path):
    """
        Saves a numpy matrix or PIL image as an image
    Args:
        im_as_arr (Numpy array): Matrix of shape DxWxH
        path (str): Path to the image
    """
    if isinstance(im, (np.ndarray, np.generic)):
        im = format_np_output(im)
        im = Image.fromarray(im)
    im.save(path)


def preprocess_image(pil_im, resize_im=True):
    """
        Processes image for CNNs

    Args:
        PIL_img (PIL_img): PIL Image or numpy array to process
        resize_im (bool): Resize to 224 or not
    returns:
        im_as_var (torch variable): Variable that contains processed float tensor
    """
    # mean and std list for channels (Imagenet)
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]

    # ensure or transform incoming image to PIL image
    if type(pil_im) != Image.Image:
        try:
            pil_im = Image.fromarray(pil_im)
        except Exception as e:
            print("could not transform PIL_img to a PIL Image object. Please check input.")

    # Resize image
    if resize_im:
        pil_im = pil_im.resize((224, 224), Image.ANTIALIAS)

    im_as_arr = np.float32(pil_im)
    im_as_arr = im_as_arr.transpose(2, 0, 1)  # Convert array to D,W,H
    # Normalize the channels
    for channel, _ in enumerate(im_as_arr):
        im_as_arr[channel] /= 255
        im_as_arr[channel] -= mean[channel]
        im_as_arr[channel] /= std[channel]
    # Convert to float tensor
    im_as_ten = torch.from_numpy(im_as_arr).float()
    # Add one more channel to the beginning. Tensor shape = 1,3,224,224
    im_as_ten.unsqueeze_(0)
    # Convert to Pytorch variable
    im_as_var = Variable(im_as_ten, requires_grad=True)
    return im_as_var


def recreate_image(im_as_var):
    """
        Recreates images from a torch variable, sort of reverse preprocessing
    Args:
        im_as_var (torch variable): Image to recreate
    returns:
        recreated_im (numpy arr): Recreated image in array
    """
    reverse_mean = [-0.485, -0.456, -0.406]
    reverse_std = [1 / 0.229, 1 / 0.224, 1 / 0.225]
    recreated_im = copy.copy(im_as_var.data.numpy()[0])
    for c in range(3):
        recreated_im[c] /= reverse_std[c]
        recreated_im[c] -= reverse_mean[c]
    recreated_im[recreated_im > 1] = 1
    recreated_im[recreated_im < 0] = 0
    recreated_im = np.round(recreated_im * 255)

    recreated_im = np.uint8(recreated_im).transpose(1, 2, 0)
    return recreated_im


def choose_best_window(data, fs=500, w_size=300):
    """
    data: array
        Must be of size k x n_samples. k can be sensor dimension or trial dimension.
    w_size: int
        The size of the window in ms
    """
    # Generate masks over time samples where sample is superior to 2x std of the vector
    masks = [dat >= (np.mean(dat) + np.std(dat) * 2) for dat in data]
    w_size = int(w_size * fs / 1000)
    best_window_idx = Parallel(n_jobs=-1)(delayed(best_window)(mask, w_size) for mask in masks)
    return best_window_idx


def best_window(mask, w_size):
    # create an array of all the possible windows in the trial with their mask values
    windows = np.array([mask[i : i + w_size] for i in range(len(mask) - w_size)])
    # get the sum of mask values in each windows. High number indicates lots of True values
    values = [sum(window) for window in windows]
    if max(values) == 0:
        return None
    # best window is where the values are the highest
    best_window_index = np.where(values == max(values))[0]
    # below is to handle if there are multiple best windows with the same values, in which case we
    # try to check if they overlap, and keep only if they are separated.
    if len(best_window_index) > 1:
        idx_range = best_window_index[-1] - best_window_index[0]
        # duration betweeen first and last window must be w_size/2
        if idx_range <= int(w_size / 2):
            # Then best would be in the middle of all this
            best = int(len(best_window_index) / 2)
            return best_window_index[best]
        elif idx_range <= w_size:
            return (best_window_index[0], best_window_index[-1])
        else:
            best = int(len(best_window_index) / 2)
            return (best_window_index[0], best, best_window_index[-1])
    else:
        # only one window, just add it.
        return best_window_index[0]


def compute_single_sal_psd(index, trial, w_size, fs):
    if isinstance(index, Iterable):
        tmp = []
        for idx in index:
            window = trial[idx : idx + w_size]
            while len(window.shape) < 3:
                window = window[np.newaxis, :]
            tmp.append(compute_psd(window, fs=fs))
        return np.mean(tmp, axis=0)
    else:
        if index is not None:
            window = trial[index : index + w_size]
            while len(window.shape) < 3:
                window = window[np.newaxis, :]
            return compute_psd(window, fs=fs)
        else:
            return [None] * 7


def compute_saliency_based_psd(saliency, trial, w_size, fs):
    chan_data = []
    for i, chan in enumerate(saliency):
        windows_idx = choose_best_window(chan, fs, w_size)
        transformed_data = Parallel(n_jobs=1)(
            delayed(compute_single_sal_psd)(index, trial[i, j], w_size, fs)
            for j, index in enumerate(windows_idx)
        )
        chan_data.append(np.array(transformed_data).squeeze())
    return np.array(chan_data)


def get_positive_negative_saliency(gradient):
    """
        Generates positive and negative saliency maps based on the gradient
    Args:
        gradient (numpy arr): Gradient of the operation to visualize

    returns:
        pos_saliency ( )
    """
    pos_saliency = np.maximum(0, gradient) / gradient.max()
    neg_saliency = np.maximum(0, -gradient) / -gradient.min()
    return pos_saliency, neg_saliency


def get_example_params(example_index):
    """
        Gets used variables for almost all visualizations, like the image, model etc.

    Args:
        example_index (int): Image id to use from examples

    returns:
        original_image (numpy arr): Original image read from the file
        prep_img (numpy_arr): Processed image
        target_class (int): Target class for the image
        file_name_to_export (string): File name to export the visualizations
        pretrained_model(Pytorch model): Model to use for the operations
    """
    # Pick one of the examples
    example_list = (
        ("../input_images/snake.jpg", 56),
        ("../input_images/cat_dog.png", 243),
        ("../input_images/spider.png", 72),
    )
    img_path = example_list[example_index][0]
    target_class = example_list[example_index][1]
    file_name_to_export = img_path[img_path.rfind("/") + 1 : img_path.rfind(".")]
    # Read image
    original_image = Image.open(img_path).convert("RGB")
    # Process image
    prep_img = preprocess_image(original_image)
    # Define model
    pretrained_model = models.alexnet(pretrained=True)
    return (
        original_image,
        prep_img,
        target_class,
        file_name_to_export,
        pretrained_model,
    )


def make_gif(image_list, output=None, duration=100, loop=0):
    if output is None:
        output = ".".join(image_list[0].split(".")[:-1] + ["gif"])
    frames = [Image.open(image) for image in image_list]
    frame_one = frames[0]
    frame_one.save(
        output,
        format="GIF",
        append_images=frames,
        save_all=True,
        duration=duration,
        loop=loop,
    )


def avg_range(arr: list):
    length = int(len(arr) / 2)
    return rec_avg_range(list(arr), length)


def rec_avg_range(arr: list, length: int):
    if len(arr) <= 1:
        return 0
    argmax = np.argmax(arr)
    max = arr.pop(argmax)
    argmin = np.argmin(arr)
    min = arr.pop(argmin)
    range = (max - min) / length
    return range + rec_avg_range(arr, length)


def _unit_test_avg_range():
    liste = [18, -8, -6, 3, -10, 15, 0]
    assert avg_range(liste) == 20


def generate_saliency_figure(
    saliencies: dict,
    info_path: str,
    save_path: str = "",
    suffix: str = "",
    title: str = "",
    sensors: list = [""],
    clf_type="",
    datatype="rest",
    sfreq=500,
    edge=50,
    cmap="coolwarm",
    stim_tick=75,
):
    """
    Generates a figure visualizing saliency maps for MEG data.

    This function creates a grid of images showing the saliency maps for
    types of stimuli (e.g., image and sound) and sensor channels (e.g., MAG, GRAD1, GRAD2).
    It also plots a topomap for the maximum saliency index along with a color bar.

    Parameters
    ----------
    saliencies : dict
        Dictionary containing saliency maps. Keys should correspond to different types
        of stimuli (e.g., "image", "sound"), and values should be numpy arrays of shape
        3 x sensors x samples, representing the saliency maps for each channel.
    save_path : str, optional
        Path to save the generated figure. Default is an empty string, which means the
        figure will not be saved automatically.
    suffix : str, optional
        Suffix to append to the filename when saving the figure. Default is an empty string.
    title : str, optional
        Title for the figure. Default is an empty string.
    sensors : list, optional
        List of sensor types to include in the visualization. Default is [""].
    clf_type : str, optional
        Flag indicating whether the saliency maps are for event classification if set to "clf".
        the y-axis ticks and labels are adjusted accordingly.
    sfreq : int, optional
        Sampling frequency for computation of  xticks. Default is  500.
    cmap : str, optional
        Colormap to use for displaying the topo- and saliency maps. Default is "coolwarm".
    stim_tick : int, optional
        Tick position for the stimulus event in the time axis. Default is  75.

    Returns
    -------
    None

    Notes
    -----
    The function assumes that the input saliency maps are normalized to have zero mean and
    unit variance. The colormap used for displaying the saliency maps is "coolwarm" by default,
    but can be changed with the `cmap` parameter.

    The function creates a grid layout with a subplot for each sensor channel and a subplot for the
    topomap. The grid layout is dynamically adjusted based on the number of sensors.

    The "eventclf" flag allows for adjustments in the y-axis ticks and labels for event classification
    scenarios, where the stimulus event is marked with a specific tick position.

    The function does not handle exceptions that may occur during the plotting process, such as issues with
    file I/O or invalid input data.
    """

    stim_tick -= int(edge / 2)
    if suffix != "" and not suffix.endswith("_"):
        suffix += "_"
    n_blocs = len(sensors)  # number of blocs of figures in a line
    n_lines = len(saliencies)  # number of lines for the pyplot figure
    n_cols = n_blocs * 3 + 1  # number of columns for the pyplot figure
    grid = GridSpec(n_lines, n_cols)
    fig = plt.figure(figsize=(n_cols * 2, n_lines * 2))
    plt.title(title)
    plt.axis("off")
    axes = []
    tick_ratio = int(1000 / sfreq)
    for i, label in enumerate(saliencies.keys()):
        gradient = saliencies[label].squeeze()
        assert (
            len(gradient) == n_blocs
        ), "Can't generate figures for all sensors, check if the saliencies have been properly computed."
        gradient /= np.abs(gradient).max()
        for j, sensor_type in zip(range(0, n_blocs * 3, n_blocs), sensors):
            idx = j // 3
            # grads = gradient[idx]
            # In an attempt to remove the edge effect:
            # We remove the first and last edge points -> therefore tick is moved to 25 (was 75)
            grads = gradient[idx][:, edge:-edge]

            segment_length = grads.shape[1]
            # We add the mid_slice variable in an attempt to tackle the edge effects by removing mean from center values for example
            # But it was uneffective. This could still be useful so we leave it here...
            mid_slice = (0, segment_length)
            # mid_slice = (int(segment_length / 4), int(3 * segment_length / 4))
            gradmeans = grads[:, mid_slice[0] : mid_slice[1]].mean(axis=1)[:, np.newaxis]
            grads -= gradmeans  # We remove mean accross time to make the variations accross time pop-up more
            n_sensors = grads.shape[0]
            vmax = grads.max()
            vmin = grads.min()
            max_idx = np.argmax([avg_range(arr) for arr in grads.T])
            # max_idx = np.argmax(np.mean(grads, axis=0))
            axes.append(fig.add_subplot(grid[i, j : j + 2]))
            plt.imshow(
                grads,
                interpolation="nearest",
                aspect=1,
                vmin=vmin,
                vmax=vmax,
                cmap=cmap,
            )
            axes[-1].spines["top"].set_visible(False)
            axes[-1].spines["right"].set_visible(False)
            axes[-1].yaxis.tick_right()

            if clf_type == "eventclf":
                x_ticks = sorted(
                    [0, stim_tick, int(segment_length / 2), segment_length] + [max_idx]
                )
                ticks_values = [(x_tick - stim_tick) * tick_ratio for x_tick in x_ticks]
                plt.axvline(x=max_idx, color="green", linestyle="--", linewidth=1)
                plt.axvline(x=stim_tick, color="black", linestyle="--", linewidth=1)
            else:
                x_ticks = [0, int(segment_length / tick_ratio), segment_length]
                ticks_values = [x_tick * tick_ratio for x_tick in x_ticks]
            plt.xticks(x_ticks, ticks_values, fontsize=8)
            plt.yticks([0, n_sensors], [n_sensors, 0])

            if j == 0:
                axes[-1].text(-50, 50, label, ha="left", va="center", rotation="vertical")
            if idx == n_blocs - 1:
                axes[-1].yaxis.set_label_position("right")
                plt.ylabel("sensors")
            if i == 0:
                plt.title(sensor_type)
            if i == 1:
                plt.xlabel("time (ms)")
            axes.append(fig.add_subplot(grid[i, j + 2]))
            if clf_type == "eventclf":
                data = grads[:, max_idx]
            else:
                data = gradmeans
            info = load_info(info_path, datatype)

            im, _ = mne.viz.plot_topomap(
                data.ravel(),
                info,
                res=300,
                cmap=cmap,
                vlim=(vmin, vmax) if vmax > vmin else (None, None),
                show=False,
                contours=0,
                axes=axes[-1],
            )
            if idx == n_blocs - 1:
                axes.append(fig.add_subplot(grid[i, n_blocs * 3]))
                fig.colorbar(
                    im,
                    ax=axes[-1],
                    location="right",
                    shrink=0.9,
                    ticks=(vmin, 0, vmax),
                )
                axes[-1].axis("off")

    out_path = os.path.join(save_path, f"{suffix}saliencies.png")
    # plt.tight_layout()
    plt.savefig(out_path, dpi=300)
    plt.close()
    return out_path


def colorFader(c1, c2, mix=0, alpha=0.5):
    # fade (linear interpolate) from color c1 (at mix=0) to c2 (mix=1)
    c1 = np.array(mpl.colors.to_rgb(c1))
    c2 = np.array(mpl.colors.to_rgb(c2))
    return mpl.colors.to_hex((1 - mix) * c1 + mix * c2)


def plot_masked_epoch(data, mask, c1="white", c2="green", alpha=0.3, title: str = None):
    dim = len(data.shape)
    assert data.shape == mask.shape, "Shape of the data should match shape of the mask."
    assert (
        dim == 2
    ), f"Data is supposed to be a 2 dimension array, a {dim}-dimension array was given."

    mask = (mask - mask.min()) / mask.ptp()
    fig, axes = plot_epoch(data, title)
    for i, ax in enumerate(axes):
        for j, x in enumerate(range(data.shape[1])):
            ax.axvline(
                x + 1,
                color=colorFader(c1, c2, mask[i, j]),
                zorder=0,
                alpha=alpha,
                linewidth=50,
            )

    return fig, axes


def plot_epoch(data, title: str = None):
    dim = len(data.shape)
    assert (
        dim == 2
    ), f"Data is supposed to be a 2 dimension array, a {dim}-dimension array was given."

    grid = GridSpec(len(data), 1, hspace=0)
    fig = plt.figure(figsize=(10, 5))
    plt.title(title)
    plt.axis("off")
    axes = []
    for i, chan in enumerate(data):
        ax = fig.add_subplot(grid[i, 0])
        plt.plot(chan, color="black", linewidth=1)
        plt.xlim(0, data.shape[1])
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        # plt.ylabel(i, rotation="horizontal")
        ax.spines["left"].set_visible(False)
        ax.set_yticks([])
        if i != len(data) - 1:
            ax.spines["bottom"].set_visible(False)
            ax.set_xticks([])
        axes.append(ax)
    plt.tight_layout()

    # out_path = os.path.join(save_path, "visualizations", "average_trial.png")
    # plt.savefig(out_path, dpi=300)
    # return out_path
    return fig, axes


def load_info(data_path, datatype):
    # Chargement des données de potition des capteurs:
    camcan_path = os.path.join(
        data_path,
        "cc700/meg/pipeline/release005/BIDSsep/",
        f"derivatives_{datatype}",
        "aa/AA_movecomp_transdef/aamod_meg_maxfilt_00003/",
    )
    subjects = os.listdir(camcan_path)
    subj_path = os.path.join(camcan_path, subjects[0])
    for file in os.listdir(subj_path):
        if file.endswith(".fif"):
            raw = mne.io.read_raw_fif(os.path.join(subj_path, file), preload=False)
            break
    # TODO check if we are correct to do this for grad topoplots, maybe use planar
    # (planar1 doesnt work and for grad they do a different computation)
    return raw.pick_types(meg="mag").info



import argparse

parser = argparse.ArgumentParser()
parser.add_argument(
    "-c",
    "--cores",
    default=-1,
    type=int,
    help="The number of cores to use, default=-1 (all cores)",
)
parser.add_argument(
    "-v",
    "--verbose",
    type=int,
    default=1,
    help="display more info must be int from 0 to 3, default=1",
)
parser.add_argument(
    "--clf",
    choices=["XGBoost", "SVM", "LDA", "QDA", "RF", "perceptron"],
    default="LDA",
    help="The classifier that will be used for the classification, default=LDA",
)
parser.add_argument(
    "-p",
    "--permutations",
    type=int,
    default=None,
    help="The number of permutations, default=None",
)
# parser.add_argument(
#     "-d",
#     "--data_type",
#     choices=["task", "rest", "passive"],
#     default="rest",
#     help="The type of data to use for classification",
# )
# parser.add_argument(
#     "--clean_type",
#     choices=["mf", "transdef_mf", "raw"],
#     default="transdef_mf",
#     help="The type of preprocessing step to use for classification",
# )
parser.add_argument(
    "-l",
    "--label",
    choices=["sex", "age", "subject"],
    default="sex",
    help="The type of classification to run, default=sex",
)
parser.add_argument(
    "-e", "--elec", default="MAG", help="The type of electrodes to keep, default=MAG"
)
parser.add_argument(
    "-f",
    "--feature",
    choices=["bands", "bins"],
    default="bands",
    help="The type of features to use, default=bands",
)
parser.add_argument(
    "--n_crossval",
    type=int,
    default=1000,
    help="The number of cross-validations to do, default=1000",
)
parser.add_argument(
    "--test_size",
    type=float,
    default=0.5,
    help="The percentage of the dataset to use as test set, default=.5",
)
parser.add_argument(
    "--iterations",
    type=int,
    default=100,
    help="The number of iterations to do for random search hyperparameter optimization, default=100",
)
parser.add_argument(
    "--test",
    action="store_true",
    help="Launch the pipeline in test mode : will not save and will only do 2 iteration for each loop",
)
parser.add_argument(
    "--force-load",
    action="store_true",
    help="forces loading and printing of shapes even if classification has been done",
)
parser.add_argument(
    "-t", "--time", action="store_true", help="keeps time and prints it at the end"
)
parser.add_argument(
    "-o",
    "--out_path",
    default=".",
    help="Where to save the result matrices, data path in config.ini file + results/, by default",
)
parser.add_argument(
    "-i", "--in_path", default=".", help="Where is the data to load"
)
parser.add_argument(
    "--elec_axis",
    type=int,
    default=1,
    help="The axis of the data where the electrodes are, default=1",
)
args = parser.parse_args()
